# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from breezedb.db import *
from breezedb.table import *
from breezedb.field import *
from breezedb.element import *
from breezedb.query import run_query
from breezedb.breeze_exceptions import *
from breezedb._version import __version__



# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. module:: breeze_exceptions
    :platform: Unix, Windows
    :synopsis: BreezeDB exception handler.

.. moduleauthor:: Rafael Medina García <rafamedgar@gmail.com>
"""

import sys

class BreezeException(Exception):
    """ This exception is raised by the modules if a regular exception
        occurs.

        :arg str level: library level in which the exception occured 
             (db, table, field, element or query)
        :arg str msg: exception information
    """

    def __init__(self, level, msg):
        self.level = level
        self.msg = msg





import unittest
import os, sys, shutil

test_root = os.path.abspath(os.path.dirname(__file__))

import breezedb

database = os.path.join(test_root, 'db_temp')

class TestTable(unittest.TestCase):

    def test_table_exists_true(self):
        # Check for a table that does exist
        result = breezedb.table_exists('cities', database)
        self.assertEqual(True, result)

    def test_table_exists_false(self):
        # Check for a table that does not exist
        result = breezedb.table_exists('test', database)
        self.assertEqual(False, result)

    def test_add_table(self):
        # Add a new table to the database
        breezedb.create_table('new_table', database)

    def test_remove_table_exists(self):
        # Remove the previously created table
        breezedb.remove_table('new_table', database)

    def test_remove_table_inexistent(self):
        # Try to remove the previously created table again
        with self.assertRaises(breezedb.BreezeException):
            breezedb.remove_table('new_table', database)

    def test_get_field_list(self):
        # Get a list of fields from table 'languages'
        expected = ['name', 'cross_platform']
        result = breezedb.get_field_list('languages', database)
        self.assertEqual(expected, result)

if __name__ == "__main__":
    # Remove previous temp copy
    if os.path.isdir(os.path.join(test_root, 'db_temp')):
        shutil.rmtree(os.path.join(test_root, 'db_temp'))
    # Create temp copy of the database
    shutil.copytree(os.path.join(test_root, 'db'),
        os.path.join(test_root, 'db_temp'))
    # Begin tests
    unittest.main()




# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. module:: element
    :platform: Unix, Windows
    :synopsis: Element related operations.

.. moduleauthor:: Rafael Medina García <rafamedgar@gmail.com>
"""

import xml.etree.ElementTree as XML
import os, string
import breezedb.field as field
import breezedb.table as table
from breezedb.breeze_exceptions import BreezeException

def element_exists(element_index, field_name, table_name, database):
    """ Check whether an element exist in the field.

        :param int element_index: index of the element to check
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database

        :returns: True or false

        :raises BreezeException: supplied index is not a positive integer
    """
    if element_index < 0:
        raise BreezeException('element', 'index needs to be a positive integer')

    field_file = os.path.join(database, table_name, field_name)
    field_tree = XML.parse(field_file)
    field_root = field_tree.getroot()

    last_index = int(field_root[-1].get('index'))
    if element_index <= last_index:
        return True
    else:
        return False

def create_element_row(element_list, table_name, database):
    """ Creates a row of elements in the given table.

        Loops through the list of fields in the table and adds the
        corresponding element.

        :param element_list: list of elements to add to the table. Elements
            must appear in the order the fields are listed in the table, and
            there must be one element per field in the list, even if they are
            left blank
        :param str table_name: name of the table that will contain the row
        :param str database: path to the database

        :raises BreezeException: cannot write to database
        :raises BreezeException: not enough elements
        :raises BreezeException: too many elements
    """
    can_write = os.access(database, os.W_OK)
    if not can_write:
        raise BreezeException('element', 'cannot write to database')

    field_list = table.get_field_list(table_name, database)
    if len(element_list) != len(field_list):
        raise BreezeException('element', 'number of elements is not equal to the number of fields')

    field_tree = XML.parse(os.path.join(database, table_name, field_list[0]))
    field_root = field_tree.getroot()

    last_index = field_root[-1].get('index')
    if last_index != None:
        index = int(last_index) + 1
    else:
        index = 0

    for it, f in enumerate(field_list):
        field_file = os.path.join(database, table_name, f)
        field_tree = XML.parse(field_file)
        field_root = field_tree.getroot()

        new_element = XML.Element('element')
        new_element.set('index', str(index))
        new_element.text = str(element_list[it])
        field_root.append(new_element)

        field_tree.write(field_file)

def parse_element_content(field_type, content):
    """ Parse the content of the element depending on the type.

        :param str field_type: type of the field that contains the element
        :param str content: content to parse

        :returns: parsed content according to the type (String, Int,
            Boolean, etc)
    """
    if field_type == 'string':
        return str(content)
    elif field_type == 'int':
        return int(content)
    elif field_type == 'float' or field_type == 'double':
        # Python's float type has double precision
        return float(content)
    elif field_type == 'boolean':
        # boolean type is represented by a 0 (false) or a 1 (true)
        if content == '0':
            return False
        else:
            return True

def get_element_content(element_index, field_name, table_name, database,
    parse = True):
    """ Get the data contained in an element.

        :param int element_index: index of the element
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database
        :param Boolean parse: whether to parse the content according to
            the type of the field or not

        :returns str: content of the element
    """
    if not element_exists(element_index, field_name, table_name, database):
        raise BreezeException('element', 'the element does not exist')

    field_file = os.path.join(database, table_name, field_name)
    field_tree = XML.parse(field_file)
    field_root = field_tree.getroot()

    if parse:
        field_type = field.get_field_type(field_name, table_name, database)
        content = field_root.findall('element')[element_index].text
        return parse_element_content(field_type, content)     

    return field_root.findall('element')[element_index].text

def find_element(to_find, field_name, table_name, database,
        ignore_case = True):
    """ Find elements in the field.

        :param str to_find: data to find
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database
        :param Boolean ignore_case: whether or not to ignore the case
            when searching

        :returns: list of indexes that match the criteria
    """
    indexlist = []

    field_file = os.path.join(database, table_name, field_name)
    field_tree = XML.parse(field_file)
    field_root = field_tree.getroot()

    for element in field_root.iter('element'):
        if ignore_case:
            if string.lower(str(to_find)) in string.lower(str(element.text)):
                indexlist.append(element.get('index'))
        else:
            if str(to_find) in str(element.text):
                indexlist.append(element.get('index'))

    return indexlist

def modify_element(element_index, field_name,
                    table_name, database, new_content):
    """ Modify the content of an element.

        :param int element_index: index of the element to modify
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database
        :param str new_content: new content to store in the element

        :raises BreezeException: element does not exist
    """
    if not element_exists(element_index, field_name, table_name, database):
        raise BreezeException('element', 'the element does not exist')

    field_file = os.path.join(database, table_name, field_name)
    field_tree = XML.parse(field_file)
    field_root = field_tree.getroot()

    field_root.findall('element')[element_index].text = str(new_content)

    field_tree.write(field_file)

def empty_element(element_index, field_name, table_name, database):
    """ Empty the content of an element.

        The content is emptied instead of removing the whole element because
        a single element cannot be removed, as the indexes of other fields
        would not match.

        :param int element_index: index of the element to empty
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database

        :raises BreezeException: element does not exist
    """
    if not element_exists(element_index, field_name, table_name, database):
        raise BreezeException('element', 'the element does not exist')

    field_file = os.path.join(database, table_name, field_name)
    field_tree = XML.parse(field_file)
    field_root = field_tree.getroot()

    field_root.findall('element')[element_index].text = ""

    field_tree.write(field_file)

def remove_element_row(element_index, table_name, database):
    """ Remove an element row from the specified table.

        Removes all the elements that correspond to that index and then
        updates the rest of the indexes of the files.

        :param int element_index: index of the element row to remove
        :param str field_name: name of the field that contains the element
        :param str table_name: name of the table that contains the field
        :param str database: path to the database

        :raises BreezeException: element does not exist,
            element row cannot be removed
    """
    table_file = os.path.join(database, table_name, 'tableinfo.breeze')
    table_tree = XML.parse(table_file)
    table_root = table_tree.getroot()

    for field in table_root:

        if not element_exists(element_index, field.text, table_name, database):
            raise BreezeException('element', 'the element does not exist')

        try:
            field_file = os.path.join(database, table_name, field.text)
            field_tree = XML.parse(field_file)
            field_root = field_tree.getroot()        

            field_root.remove(field_root.findall('element')[element_index])

            for element in field_root.iter('element'):
                if int(element.get('index')) > element_index:
                    element.set('index', str(int(element.get('index')) - 1))

            field_tree.write(field_file)

        except OSError:
            raise BreezeException('element', 'could not remove element')




# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. module:: query
    :platform: Unix, Windows
    :synopsis: Query parsing for breezedb.

.. moduleauthor:: Rafael Medina García <rafamedgar@gmail.com>
"""

import breezedb, re
from breezedb.breeze_exceptions import BreezeException

# Regular expression for arguments
RE_ARG = re.compile("\'(.*?)\'")

class Parser():
    """ Parses the query and divides it into subqueries where possible.

        :arg str query: query to parse
    """

    def __init__(self, query):
        self.query = query

    def run(self):
        """ Run the query. """
        # Check the type of operation
        if re.match("CREATE (.*)", self.query):
            self.create()
        elif re.match("GET (.*)", self.query):
            return self.get()
        elif re.match("REMOVE (.*)", self.query):
            self.remove()
        elif re.match("RENAME (.*)", self.query):
            self.rename()
        elif re.match("EXISTS (.*)", self.query):
            return self.exists()
        elif re.match("EMPTY (.*)", self.query):
            self.empty()
        elif re.match("FIND (.*)", self.query):
            return self.find()
        elif re.match("MODIFY (.*)", self.query):
            self.modify()
        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)
            
    def create(self):
        """ Run a CREATE operation. This operation works with databases,
            tables and fields.

            :raises BreezeException: incorrect query syntax           
        """
        re_create_db = re.compile("CREATE DB \'(.*?)\' AT \'(.*?)\'")
        re_create_table = re.compile("CREATE TABLE (.*) AT \'(.*?)\'")
        re_create_field = re.compile("CREATE FIELD (.*) IN \'(.*?)\' AT \'(.*?)\'")
        re_create_row = re.compile("CREATE ROW (.*) IN \'(.*?)\' AT \'(.*?)\'")

        if re_create_db.match(self.query):
            # CREATE DB 'name' AT 'path'
            db_name = re_create_db.match(self.query).group(1)
            db_path = re_create_db.match(self.query).group(2)
            breezedb.create_db(db_path, db_name)
        
        elif re_create_table.match(self.query):
            # CREATE TABLE 'table1' 'table2' ...  AT 'db'
            table_args = re_create_table.match(self.query).group(1)
            db_path = re_create_table.match(self.query).group(2)
            table_list = RE_ARG.findall(table_args)

            for table in table_list:
                breezedb.create_table(table, db_path)

        elif re_create_field.match(self.query):
            # CREATE FIELD 'name1' 'type1' 'name2' 'type2' ... IN 'table' AT 'db'
            field_args = re_create_field.match(self.query).group(1)
            table_name = re_create_field.match(self.query).group(2)
            db_path = re_create_field.match(self.query).group(3)
            field_list = RE_ARG.findall(field_args)

            it = 0
            while it < len(field_list):
                    breezedb.create_field(field_list[it], field_list[it+1],
                            table_name, db_path)
                    it += 2

        elif re_create_row.match(self.query):
            # CREATE ROW 'element' 'element' ... IN 'table' AT 'db'
            element_args = re_create_row.match(self.query).group(1)
            table_name = re_create_row.match(self.query).group(2)
            db_path = re_create_row.match(self.query).group(3)
            element_list = RE_ARG.findall(element_args)

            breezedb.create_element_row(element_list, table_name, db_path)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def get(self):
        """ Run a GET operation. This operation works with databases,
            tables, fields and elements.

            :returns: list of results obtained from the query

            :raises BreezeException: incorrect query syntax 
        """
        re_get_tables = re.compile("GET TABLES AT \'(.*?)\'")
        re_get_fields = re.compile("GET FIELDS IN \'(.*?)\' AT \'(.*?)\'")
        re_get_type = re.compile("GET TYPE OF \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")
        re_get_elements = re.compile("GET ELEMENTS FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")
        re_get_row = re.compile("GET ROW \'(\d)\' IN \'(.*?)\' AT \'(.*?)\'")
        re_get_element = re.compile("GET ELEMENT \'(\d)\' FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")

        if re_get_tables.match(self.query):
            # GET TABLES AT 'db'
            db_path = re_get_tables.match(self.query).group(1)

            return breezedb.get_table_list(db_path)

        elif re_get_fields.match(self.query):
            # GET FIELDS IN 'table' AT 'db'
            table_name = re_get_fields.match(self.query).group(1)
            db_path = re_get_fields.match(self.query).group(2)
            
            return breezedb.get_field_list(table_name, db_path)

        elif re_get_type.match(self.query):
            # GET TYPE OF 'field' IN 'table' AT 'db'
            field_name = re_get_type.match(self.query).group(1)
            table_name = re_get_type.match(self.query).group(2)
            db_path = re_get_type.match(self.query).group(3)
            
            return breezedb.get_field_type(field_name, table_name, db_path)

        elif re_get_elements.match(self.query): 
            # GET ELEMENTS FROM 'field' IN 'table' AT 'db'
            field_name = re_get_elements.match(self.query).group(1)
            table_name = re_get_elements.match(self.query).group(2)
            db_path = re_get_elements.match(self.query).group(3)
            
            return breezedb.get_element_list(field_name, table_name, db_path)

        elif re_get_row.match(self.query):
            # GET ROW 'index' IN 'table' AT 'db'
            index = int(re_get_row.match(self.query).group(1))
            table_name = re_get_row.match(self.query).group(2)
            db_path = re_get_row.match(self.query).group(3)

            return breezedb.get_element_row(index, table_name, db_path)

        elif re_get_element.match(self.query):
            # GET ELEMENT 'index' FROM 'field' IN 'table' AT 'db'
            index = int(re_get_element.match(self.query).group(1))
            field_name = re_get_element.match(self.query).group(2)
            table_name = re_get_element.match(self.query).group(3)
            db_path = re_get_element.match(self.query).group(4)

            return breezedb.get_element_content(index, field_name,
                    table_name, db_path)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def remove(self):
        """ Run a REMOVE operation. This operation works with databases,
            tables, fields and elements.

            :raises BreezeException: incorrect query syntax 
        """
        re_remove_db = re.compile("REMOVE DB AT \'(.*?)\'")
        re_remove_table = re.compile("REMOVE TABLE (.*) AT \'(.*?)\'")
        re_remove_field = re.compile("REMOVE FIELD (.*) IN \'(.*?)\' AT \'(.*?)\'")
        re_remove_row = re.compile("REMOVE ROW (.*) IN \'(.*?)\' AT \'(.*?)\'")

        if re_remove_db.match(self.query):
            # REMOVE DB AT 'path'
            db_path = re_remove_db.match(self.query).group(1)

            breezedb.remove_db(db_path)

        elif re_remove_table.match(self.query):
            # REMOVE TABLE 'table1' 'table2' ... AT 'db'
            table_args = re_remove_table.match(self.query).group(1)
            db_path = re_remove_table.match(self.query).group(2)
            table_list = RE_ARG.findall(table_args)

            for table in table_list:
                breezedb.remove_table(table, db_path)

        elif re_remove_field.match(self.query):
            # REMOVE FIELD 'field1' 'field2' ... IN 'table' AT 'db'
            field_args = re_remove_field.match(self.query).group(1)
            table_name = re_remove_field.match(self.query).group(2)
            db_path = re_remove_field.match(self.query).group(3)
            field_list = RE_ARG.findall(field_args)

            for field in field_list:
                breezedb.remove_field(field, table_name, db_path)

        elif re_remove_row.match(self.query):
            # REMOVE ROW 'index1' 'index2' ... IN 'table' AT 'db'
            index_args = re_remove_row.match(self.query).group(1)
            table_name = re_remove_row.match(self.query).group(2)
            db_path = re_remove_row.match(self.query).group(3)
            index_list = RE_ARG.findall(index_args)

            for index in index_list:
                breezedb.remove_element_row(int(index), table_name, db_path)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def rename(self):
        """ Run a RENAME operation. This operation works with tables and
            fields.

            :raises BreezeException: incorrect query syntax 
        """
        re_rename_table = re.compile("RENAME TABLE \'(.*?)\' AT \'(.*?)\' TO \'(.*?)\'")
        re_rename_field = re.compile("RENAME FIELD \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\' TO \'(.*?)\'")

        if re_rename_table.match(self.query) :
            # RENAME TABLE 'name' AT 'db' TO 'new name'
            table_name = re_rename_table.match(self.query).group(1)
            db_path = re_rename_table.match(self.query).group(2)
            new_name = re_rename_table.match(self.query).group(3)

            breezedb.rename_table(table_name, db_path, new_name)
        
        elif re_rename_field.match(self.query):
            # RENAME FIELD 'name' IN 'table' AT 'db' TO 'new name'
            field_name = re_rename_field.match(self.query).group(1)
            table_name = re_rename_field.match(self.query).group(2)
            db_path = re_rename_field.match(self.query).group(3)
            new_name = re_rename_field.match(self.query).group(4)

            breezedb.rename_field(field_name, table_name, db_path, new_name)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def exists(self):
        """ Run an EXISTS operation. This operation works with tables,
            fields and elements.

            :returns: True or False

            :raises BreezeException: incorrect query syntax 
        """
        re_exists_table = re.compile("EXISTS TABLE \'(.*?)\' AT \'(.*?)\'")
        re_exists_field = re.compile("EXISTS FIELD \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")
        re_exists_element = re.compile("EXISTS ELEMENT \'(\d)\' FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")

        if re_exists_table.match(self.query):
            # EXISTS TABLE 'table' AT 'db'
            table_name = re_exists_table.match(self.query).group(1)
            db_path = re_exists_table.match(self.query).group(2)

            return breezedb.table_exists(table_name, db_path)

        if re_exists_field.match(self.query):
            # EXISTS FIELD 'field' IN 'table' AT 'db'
            field_name = re_exists_field.match(self.query).group(1)
            table_name = re_exists_field.match(self.query).group(2)
            db_path = re_exists_field.match(self.query).group(3)

            return breezedb.field_exists(field_name, table_name, db_path)

        if re_exists_element.match(self.query):
            # EXISTS ELEMENT 'index' FROM 'field' IN 'table' AT 'db'
            index = int(re_exists_element.match(self.query).group(1))
            field_name = re_exists_element.match(self.query).group(2)
            table_name = re_exists_element.match(self.query).group(3)
            db_path = re_exists_element.match(self.query).group(4)

            return breezedb.element_exists(index, field_name, table_name,
                    db_path)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def empty(self):
        """ Run an EMPTY operation. This operation works with fields and
            elements.

            :raises BreezeException: incorrect query syntax 
        """
        re_empty_field = re.compile("EMPTY FIELD (.*) IN \'(.*?)\' AT \'(.*?)\'")
        re_empty_element = re.compile("EMPTY ELEMENT (.*) FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")

        if re_empty_field.match(self.query):
            # EMPTY FIELD 'field1' 'field2' ... IN 'table' AT 'db'
            field_args = re_empty_field.match(self.query).group(1)
            table_name = re_empty_field.match(self.query).group(2)
            db_path = re_empty_field.match(self.query).group(3)
            field_list = RE_ARG.findall(field_args)

            for field in field_list:
                breezedb.empty_field(field, table_name, db_path)

        elif re_empty_element.match(self.query):
            # EMPTY ELEMENT 'index1' 'index2' ... FROM 'field' IN 'table' AT 'db'
            index_args = re_empty_element.match(self.query).group(1)
            field_name = re_empty_element.match(self.query).group(2)
            table_name = re_empty_element.match(self.query).group(3)
            db_path = re_empty_element.match(self.query).group(4)
            index_list = RE_ARG.findall(index_args)

            for index in index_list:
                breezedb.empty_element(int(index), field, table, database)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def find(self):
        """ Run a FIND operation. This operation only works with elements.

            :returns: list of results obtained from the query

            :raises BreezeException: incorrect query syntax 
        """
        re_find = re.compile("FIND \'(.*?)\' FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\'")
 
        if re_find.match(self.query):
            # FIND 'content' FROM 'field' IN 'table' AT 'db'
            content = re_find.match(self.query).group(1)
            field_name = re_find.match(self.query).group(2)
            table_name = re_find.match(self.query).group(3)
            db_path = re_find.match(self.query).group(4)

            return breezedb.find_element(content, field_name, table_name, db_path)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

    def modify(self):
        """ Run a MODIFY operation. This operation only works with elements. 

            :raises BreezeException: incorrect query syntax 
        """
        re_modify = re.compile("MODIFY \'(\d)\' FROM \'(.*?)\' IN \'(.*?)\' AT \'(.*?)\' TO \'(.*?)\'")

        if re_modify.match(self.query):
            # MODIFY 'index' FROM 'field' IN 'table' AT 'db' TO 'new content'
            index = int(re_modify.match(self.query).group(1))
            field_name = re_modify.match(self.query).group(2)
            table_name = re_modify.match(self.query).group(3)
            db_path = re_modify.match(self.query).group(4)
            new_content = re_modify.match(self.query).group(5)

            breezedb.modify_element(index, field, table, database, new_content)

        else:
            raise BreezeException('query', 'invalid query: %s' % self.query)

def run_query(query):
    """ Parse and execute a query in the database.
        
        This function divides the query (if there are more than one) by
        using the string ';;' and then runs a parser for each one.
    
        :param str query: query to execute
    """
    query_list = query.split(';;')
    result_list = []
    for subquery in query_list:
        parser = Parser(subquery)
        result = parser.run()
        if result:
            result_list.append(result)
    
    # Results must be parsed manually
    if result_list:
        return result_list




# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. module:: table
    :platform: Unix, Windows
    :synopsis: Table related operations.

.. moduleauthor:: Rafael Medina García <rafamedgar@gmail.com>
"""

import xml.etree.ElementTree as XML
import os, shutil, breezedb
from breezedb.breeze_exceptions import BreezeException

def table_exists(table_name, database):
    """ Check whether a table exists in the database or not.
    
        :param table_name: name of the table to check
        :type table_name: str
        :param database: path to the database in which to check
        :type database: str

        :returns: True or False
    """
    breeze_file = os.path.join(database, 'root.breeze')
    breeze_tree = XML.parse(breeze_file)
    breeze_root = breeze_tree.getroot()   
 
    exists = False
    for table in breeze_root:
        if table.text == table_name:
            exists = True
            break

    if not exists:
        return False

    table_path = os.path.join(database, table_name)
    if os.path.exists(table_path):
        exists = True
    else:
        exists  = False

    if not exists:
        return False

    table_file = os.path.join(table_path, 'tableinfo.breeze')
    if os.path.isfile(table_file):
        return True
    else:
        return False

def create_table(table_name, database):
    """ Create a new table in the database.

        Adds a new <table> element to the root.breeze file and create the
        corresponding table structure.

        :param str table_name: name of the new table
        :param str database: path to the database

        :raises BreezeException: database is not writable,
            table already exists, table cannot be created
    """
    can_write = os.access(database, os.W_OK)
    if not can_write:
        raise BreezeException('table', 'cannot write to database')

    if table_exists(table_name, database):
        raise BreezeException('table', 'table already exists')

    try:
        newdir = os.path.join(database, table_name)
        os.makedirs(newdir, 0755)

        table_file = os.path.join(newdir, 'tableinfo.breeze')
        breeze_tag = XML.Element('breeze')
        table_tree = XML.ElementTree(breeze_tag)
        table_tree.write(table_file)

        breeze_file = os.path.join(database, 'root.breeze')
        breeze_tree = XML.parse(breeze_file)
        breeze_root = breeze_tree.getroot()

        new_table = XML.Element('table')
        new_table.text = table_name
        breeze_root.append(new_table)

        breeze_tree.write(breeze_file)

    except OSError:
        raise BreezeException('table', 'could not create base directory')

    except IOError:
        raise BreezeException('table', 'error writing to file')

def rename_table(table_name, database, new_name):
    """ Rename a table from the database.

        :param str table_name: current name of the table
        :param str database: path to the database
        :param str new_name: new name for the table

        :raises BreezeException: database is not writable,
            the specified table does not exist,
            a table named `new_name` already exists
    """
    can_write = os.access(database, os.W_OK)
    if not can_write:
        raise BreezeException('table', 'cannot write to database')

    if not table_exists(table_name, database):
        raise BreezeException('table', 'table does not exist')

    if table_exists(new_name, database):
        raise BreezeException('table', 'table %s already exists'
            % new_name)

    try:
        breeze_file = os.path.join(database, 'root.breeze')
        breeze_tree = XML.parse(breeze_file)
        breeze_root = breeze_tree.getroot()

        for table in breeze_root:
            if table.text == table_name:
                table.text = new_name
                break

        breeze_tree.write(breeze_file)

        src = os.path.join(database, table_name)
        dst = os.path.join(database, new_name)
        os.rename(src, dst)

    except IOError:
        raise BreezeException('table', 'could not rename table')

    except OSError:
        raise BreezeException('table', 'cold not rename directory')

def remove_table(table_name, database):
    """ Remove a table from the database.

        Removes the corresponding directory and <table> element
        from root.breeze file

        :param str table_name: name of the table to remove
        :param str database: path to the database

        :raises BreezeException: database is not writable,
            table does not exist
    """
    can_write = os.access(database, os.W_OK)
    if not can_write:
        raise BreezeException('table', 'cannot write to database')

    if not table_exists(table_name, database):
        raise BreezeException('table', 'table does not exist')

    try:
        breeze_file = os.path.join(database, 'root.breeze')
        breeze_tree = XML.parse(breeze_file)
        breeze_root = breeze_tree.getroot()

        for table in breeze_root:
            if table.text == table_name:
                breeze_root.remove(table)
                break

        breeze_tree.write(breeze_file)

        table_dir = os.path.join(database, table_name)
        shutil.rmtree(table_dir)

    except OSError:
        raise BreezeException('table', 'could not remove table directory')

    except IOError:
        raise BreezeException('table', 'error writing to file')

def get_field_list(table_name, database):
    """ Get a list of fields present in the table.

        :param str table_name: name of the table from which to get
            the fields
        :param str database: path to the database

        :raises BreezeException: table does not exist
    """
    if not table_exists(table_name, database):
        print table_name
        raise BreezeException('table', 'table does not exist')

    fieldlist = []

    table_file = os.path.join(database, table_name, 'tableinfo.breeze')
    table_tree = XML.parse(table_file)
    table_root = table_tree.getroot()

    for field in table_root:
        fieldlist.append(field.text)

    return fieldlist

def get_element_row(index, table_name, database):
    """ Get the elements located in a row of the table.

        :param int index: index of the row
        :param str table_name: name of the table from which to get the
            elements
        :param str database: path to the database
    """
    elementlist = []

    table_file = os.path.join(database, table_name, 'tableinfo.breeze')
    table_tree = XML.parse(table_file)
    table_root = table_tree.getroot()

    for field in table_root:
        elementlist.append(breezedb.get_element_content(index, field.text,
            table_name, database))

    return elementlist




# -*- coding: utf-8 -*-
#
# This file is part of breezedb - https://github.com/RMed/breezedb_python
#
# Copyright (C) 2013  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. module:: db
    :platform: Unix, Windows
    :synopsis: DB related operations.

.. moduleauthor:: Rafael Medina García <rafamedgar@gmail.com>
"""

import xml.etree.ElementTree as XML
import os, shutil
from breezedb.breeze_exceptions import BreezeException

def get_table_list(database):
    """ Get a list of tables present in the database's root.breeze file.

        :param str database: path to the database

        :returns: list of all the tables

        :raises BreezeException: directory is not a database
    """
    breeze_file = os.path.join(database, 'root.breeze')
    is_breezedb = os.path.isfile(breeze_file)

    if not is_breezedb:
        raise BreezeException('db', 'not a database')

    tablelist = []

    breeze_file = os.path.join(database, 'root.breeze')
    breeze_tree = XML.parse(breeze_file)
    breeze_root = breeze_tree.getroot()

    for table in breeze_root:
        tablelist.append(table.text)

    return tablelist
        
def create_db(path, name):
    """ Create a database structure in the specified path.

        :param str path: path where the database should be created
        :param str name: name for the database

        :raises BreezeException: cannot write to path,
            cannot create directory, cannot create root.breeze file
    """
    can_write = os.access(path, os.W_OK)
    if not can_write:
        raise BreezeException('db', 'cannot write to path')

    try:
        newdir = os.path.join(path, name)
        os.makedirs(newdir)

        breeze_file = os.path.join(path, name, 'root.breeze')
        breeze_tag = XML.Element('breeze')
        breeze_tree = XML.ElementTree(breeze_tag)
        breeze_tree.write(breeze_file)

    except OSError:
        raise BreezeException('db', 'could not create base directory')

    except IOError:
        raise BreezeException('db', 'error creating the root.breeze file')

def remove_db(path):
    """ Remove the breeze directory structure of the specified path.

        :param str path: Path to the database

        :raises BreezeException: cannot write to path,
            cannot remove the directory
    """
    can_write = os.access(path, os.W_OK)
    breeze_file = os.path.join(path, 'root.breeze')
    is_breezedb = os.path.isfile(breeze_file)

    if not can_write or not is_breezedb:
        raise BreezeException('db', 'cannot remove')

    try:
        shutil.rmtree(path)

    except:
        raise BreezeException('db', 'could not remove the database')

